/**
 * @file Makes http calls to remote resources as `Cmd`s.
 *
 * See [Http](https://package.elm-lang.org/packages/elm/http/latest/Http) Elm package.
 */
import * as E from 'fp-ts/es6/Either';
import * as O from 'fp-ts/es6/Option';
import * as T from 'fp-ts/es6/Task';
import { flow, identity } from 'fp-ts/es6/function';
import { pipe } from 'fp-ts/es6/pipeable';
import { of } from 'rxjs';
import { AjaxError, AjaxTimeoutError, ajax } from 'rxjs/ajax';
import { catchError, map } from 'rxjs/operators';
/**
 * @since 0.5.0
 */
export function toTask(req) {
    return () => xhr(req).toPromise();
}
/**
 * Executes as `Cmd` the provided call to remote resource, mapping result to a `Msg`.
 * @since 0.5.0
 */
export function send(f) {
    return req => xhr(req).pipe(map(result => T.of(O.some(f(result)))));
}
/**
 * @since 0.5.0
 */
export function get(url, decoder) {
    return {
        method: 'GET',
        headers: {},
        url,
        body: undefined,
        expect: decoder,
        timeout: O.none,
        withCredentials: false
    };
}
/**
 * @since 0.5.0
 */
export function post(url, body, decoder) {
    return {
        method: 'POST',
        headers: {},
        url,
        body,
        expect: decoder,
        timeout: O.none,
        withCredentials: false
    };
}
// --- Helpers
function xhr(req) {
    return ajax(toXHRRequest(req)).pipe(map(flow(toResponse(req), decodeWith(req.expect))), catchError((e) => of(E.left(toHttpError(e)))));
}
function toXHRRequest(req) {
    return Object.assign(Object.assign({}, req), { timeout: pipe(req.timeout, O.fold(() => 0, identity)), async: true, responseType: 'json' });
}
function toResponse(req) {
    return resp => ({
        url: req.url,
        status: {
            code: resp.status,
            message: ''
        },
        headers: req.headers,
        body: resp.response
    });
}
function decodeWith(decoder) {
    return resp => pipe(decoder(resp.body), E.mapLeft(e => ({
        _tag: 'BadPayload',
        value: e,
        response: resp
    })));
}
function toHttpError(e) {
    if (e instanceof AjaxTimeoutError) {
        return { _tag: 'Timeout' };
    }
    // RxJS ajax method can throw errors of different types (not just `AjaxError`).
    // These controls seem to cover every case.
    if (e instanceof AjaxError && e.status === 404) {
        return { _tag: 'BadUrl', value: e.request.url };
    }
    if (e instanceof AjaxError && e.status !== 0) {
        return { _tag: 'BadStatus', response: e.response };
    }
    // Parsing json in response could throw a `SyntaxError`.
    if ('constructor' in e && 'name' in e.constructor && e.constructor.name === 'SyntaxError') {
        return { _tag: 'BadPayload', value: e.message, response: e.response };
    }
    return { _tag: 'NetworkError', value: e.message };
}
